# Create your models here.
import uuid
from datetime import timedelta
from typing import Optional

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import (
    InMemoryUploadedFile,
    SimpleUploadedFile,
    TemporaryUploadedFile,
)
from django.db import models
from django.db.models import Count, Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from imagekit.models import ProcessedImageField
from pilkit.processors import ResizeToFit
from rest_framework.exceptions import ValidationError
from sortedm2m.fields import SortedManyToManyField

from comments.models import Comment
from comments.models.commentable_notification_context import (
    CommentNotificationMixin,
    CommentNotificationContext,
)
from openbook_auth.models import User, get_reactions
from openbook_common.enums import ReactionType
from openbook_common.helpers import (
    ExifRotate,
    PrefixedQ,
    attr_filled,
    generate_blurhash,
)
from openbook_common.models import (
    LinkPreview,
    ModelWithUUID,
    OrderedModelWithUUID,
)
from openbook_common.types import LanguageCode
from openbook_common.utils.helpers import (
    delete_file_field,
    extract_hashtags_from_string,
    extract_usernames_from_string,
    get_magic,
)
from openbook_common.tracking import track, TrackingEvent
from openbook_common.utils.model_loaders import (
    get_community_model,
    get_community_new_post_notification_model,
    get_community_notifications_subscription_model,
    get_hashtag_model,
    get_post_comment_notification_model,
    get_post_comment_reaction_notification_model,
    get_post_comment_reply_notification_model,
    get_post_comment_user_mention_notification_model,
    get_post_reaction_notification_model,
    get_post_user_mention_model,
    get_post_user_mention_notification_model,
    get_trending_post_model,
    get_user_model,
    get_user_new_post_notification_model,
    get_user_notifications_subscription_model,
)
from openbook_moderation.models import ModeratedObject
from openbook_notifications.services import NotificationsService
from openbook_posts.checkers import (
    check_can_add_media,
    check_can_be_published,
    check_can_be_updated,
    check_mimetype_is_supported_media_mimetypes,
)
from openbook_posts.helpers import (
    upload_to_post_image_directory,
)
from openbook_posts.jobs import process_post_media
from openbook_posts.validators import post_text_validators

magic = get_magic()
post_image_storage = default_storage


class Post(CommentNotificationMixin, ModelWithUUID):
    moderated_object = GenericRelation(ModeratedObject, related_query_name="posts")

    title = models.CharField(max_length=settings.POST_TITLE_MAX_LENGTH, blank=False, null=True)
    text = models.TextField(
        _("text"),
        max_length=settings.POST_MAX_LENGTH,
        blank=False,
        null=True,
        validators=post_text_validators,
    )
    created = models.DateTimeField(editable=False, db_index=True)
    modified = models.DateTimeField(db_index=True, default=timezone.now)
    creator = models.ForeignKey(User, on_delete=models.CASCADE, related_name="posts")
    comments_enabled = models.BooleanField(_("comments enabled"), default=True, editable=False, null=False)
    public_reactions = models.BooleanField(_("public reactions"), default=True, editable=False, null=False)
    community = models.ForeignKey(
        "openbook_communities.Community",
        on_delete=models.CASCADE,
        related_name="posts",
        null=True,
        blank=True,
    )
    language_code = models.CharField(
        max_length=2, choices=LanguageCode.choices, blank=True, null=True, verbose_name=_("language")
    )
    is_edited = models.BooleanField(default=False)
    is_closed = models.BooleanField(default=False)
    is_deleted = models.BooleanField(default=False)
    STATUS_DRAFT = "D"
    STATUS_PROCESSING = "PG"
    STATUS_PUBLISHED = "P"
    STATUSES = (
        (STATUS_DRAFT, "Draft"),
        (STATUS_PROCESSING, "Processing"),
        (STATUS_PUBLISHED, "Published"),
    )
    status = models.CharField(blank=False, null=False, choices=STATUSES, default=STATUS_DRAFT, max_length=2)
    # we need a sorted m2m for having the linkpreviews in the same order as the links in the text.
    links = SortedManyToManyField(LinkPreview, related_name="+", blank=True)
    mentions = models.ManyToManyField(User, related_name="mentioned_in_posts", blank=True, through="PostUserMention")
    is_public = models.BooleanField(default=False, db_index=True)
    comments = GenericRelation(
        "comments.Comment",
        content_type_field="commentable_type",
        object_id_field="commentable_id",
        related_query_name="post",
    )

    class Meta:
        index_together = [
            ("creator", "community"),
        ]

    @classmethod
    def post_with_id_has_public_reactions(cls, post_id):
        return Post.objects.filter(pk=post_id, public_reactions=True).exists()

    @classmethod
    def is_post_with_id_a_community_post(cls, post_id):
        return Post.objects.filter(pk=post_id, community__isnull=False).exists()

    @classmethod
    def create_post(
        cls,
        creator,
        circles_ids=None,
        community_name=None,
        community_id=None,
        image=None,
        text=None,
        video=None,
        created=None,
        is_draft=False,
        title=None,
        is_public=False,
        language=None,
    ) -> "Post":
        if not community_name and not community_id and not circles_ids and not is_public:
            raise ValidationError(_("A post requires circles or a community or to be public to be posted to."))

        if image and video:
            raise ValidationError(_("A post must have an image or a video, not both."))

        post = Post.objects.create(
            creator=creator, created=created, title=title, is_public=is_public, language_code=language
        )

        Community = get_community_model()
        if community_id:
            post.community = Community.objects.get(id=community_id)
        elif community_name:
            post.community = Community.objects.get(name=community_name)

        if text:
            post.text = text

        if image:
            post.add_media(file=image)
        elif video:
            post.add_media(file=video)

        if circles_ids:
            post.circles.add(*circles_ids)

        # If on create we have a video or image, we automatically publish it
        # Backwards compatibility reasons.
        if not is_draft:
            post.publish()
        else:
            post.save()

        track(
            creator,
            TrackingEvent(
                "spacePostCreated",
                {"spaceId": str(community_id), "postId": str(post.id)},
            ),
        )

        return post

    def update_post(
        self,
        is_public=None,
        image=None,
        text=None,
        video=None,
        is_draft=False,
        title=None,
        remove_image=False,
        language=None,
    ) -> "Post":
        if image and video:
            raise ValidationError(_("A post must have an image or a video, not both."))

        if self.is_draft() and not is_draft:
            raise ValidationError(_("You cannot change the status of a published post to draft."))

        self.title = title
        self.is_public = is_public

        if text:
            self.text = text

        if language:
            self.language_code = language

        if remove_image:
            self._delete_media_image()
        elif image:
            self._update_image(image)

        if self.is_draft() and not is_draft:
            self.publish()
        else:
            self.save()

        self.refresh_from_db()  # relation is not updated correctly in self
        return self

    @classmethod
    def get_reaction_counts_for_post_with_id(cls, post_id, reaction_type=None, reactor_id=None):
        reactions_query = Q(
            post_id=post_id,
        )

        if reaction_type:
            reactions_query.add(Q(reaction_type=reaction_type), Q.AND)

        if reactor_id:
            reactions_query.add(Q(reactor__id=reactor_id), Q.AND)

        return get_reactions(PostReaction, reactions_query)

    @classmethod
    def get_trending_posts_for_user_with_id(cls, user_id, max_id=None, min_id=None):
        """
        Gets trending posts (communities only) for authenticated user excluding reported, closed, blocked users posts
        """
        Post = cls
        TrendingPost = get_trending_post_model()
        Community = get_community_model()

        posts_select_related = (
            "post__creator",
            "post__creator__profile",
            "post__community",
            "post__image",
        )
        posts_prefetch_related = (
            "post__circles",
            "post__creator__profile__badges",
            "post__reactions__reactor",
        )

        posts_only = (
            "id",
            "post__text",
            "post__id",
            "post__created",
            "post__image__width",
            "post__image__height",
            "post__image__image",
            "post__creator__username",
            "post__creator__id",
            "post__creator__profile__name",
            "post__creator__profile__avatar",
            "post__creator__profile__badges__id",
            "post__creator__profile__badges__keyword",
            "post__creator__profile__id",
            "post__community__id",
            "post__community__name",
            "post__community__avatar",
            "post__community__color",
            "post__community__title",
        )

        trending_community_posts_query = Q(
            post__is_closed=False,
            post__is_deleted=False,
            post__status=Post.STATUS_PUBLISHED,
        )

        trending_community_posts_query.add(
            ~Q(
                Q(post__creator__blocked_by_users__blocker_id=user_id)
                | Q(post__creator__user_blocks__blocked_user_id=user_id)
            ),
            Q.AND,
        )
        trending_community_posts_query.add(Q(post__community__type=Community.COMMUNITY_TYPE_PUBLIC), Q.AND)
        trending_community_posts_query.add(~Q(post__community__banned_users__id=user_id), Q.AND)

        if max_id:
            trending_community_posts_query.add(Q(id__lt=max_id), Q.AND)
        elif min_id:
            trending_community_posts_query.add(Q(id__gt=min_id), Q.AND)

        trending_community_posts_queryset = (
            TrendingPost.objects.select_related(*posts_select_related)
            .prefetch_related(*posts_prefetch_related)
            .only(*posts_only)
            .filter(trending_community_posts_query)
        )

        return trending_community_posts_queryset

    @classmethod
    def get_trending_posts_old_for_user_with_id(cls, user_id):
        """
        For backwards compatibility reasons
        """
        trending_posts_query = cls._get_trending_posts_old_query()
        trending_posts_query.add(~Q(community__banned_users__id=user_id), Q.AND)

        trending_posts_query.add(
            ~Q(Q(creator__blocked_by_users__blocker_id=user_id) | Q(creator__user_blocks__blocked_user_id=user_id)),
            Q.AND,
        )

        trending_posts_query.add(~Q(moderated_object__status=ModeratedObject.STATUS_APPROVED), Q.AND)

        return cls._get_trending_posts_old_with_query(query=trending_posts_query)

    @classmethod
    def _get_trending_posts_old_with_query(cls, query):
        return cls.objects.filter(query).annotate(Count("reactions")).order_by("-reactions__count", "-created")

    @classmethod
    def _get_trending_posts_old_query(cls):
        trending_posts_query = Q(created__gte=timezone.now() - timedelta(hours=12))

        Community = get_community_model()

        trending_posts_sources_query = Q(
            community__type=Community.COMMUNITY_TYPE_PUBLIC,
            status=cls.STATUS_PUBLISHED,
            is_closed=False,
            is_deleted=False,
        )

        trending_posts_query.add(trending_posts_sources_query, Q.AND)

        return trending_posts_query

    @classmethod
    def get_post_comment_notification_target_users(cls, post, post_commenter):
        """
        Returns the users that should be notified of a post comment.
        This includes the post creator and other post commenters
        :return:
        """

        # Add other post commenters, exclude replies to comments, the post commenter
        other_commenters = User.objects.filter(
            Q(
                comments__commentable_id=post.pk,
                comments__parent_comment_id=None,
            )
            & ~Q(id=post_commenter.pk)
        )

        post_creator = User.objects.filter(pk=post.creator_id)

        return other_commenters.union(post_creator)

    @classmethod
    def get_post_comment_reply_notification_target_users(cls, post_commenter, parent_post_comment):
        """
        Returns the users that should be notified of a post comment reply.
        :return:
        """

        # Add other post commenters, exclude non replies, the post commenter
        other_repliers = User.objects.filter(
            Q(
                comments__parent_comment_id=parent_post_comment.pk,
            )
            & ~Q(id=post_commenter.pk)
        )

        # Add post comment creator
        post_comment_creator = User.objects.filter(pk=parent_post_comment.commenter_id)

        # Add post creator
        post = parent_post_comment.commentable
        post_creator = User.objects.filter(pk=post.creator.id)
        return other_repliers.union(post_comment_creator, post_creator)

    @classmethod
    def get_community_notification_target_subscriptions(cls, post):
        CommunityNotificationsSubscription = get_community_notifications_subscription_model()

        community_subscriptions_query = Q(community=post.community, new_post_notifications=True)

        exclude_blocked_users_query = Q(
            Q(subscriber__blocked_by_users__blocker_id=post.creator.pk)
            | Q(subscriber__user_blocks__blocked_user_id=post.creator.pk)
        )
        community_members_query = Q(subscriber__communities_memberships__community_id=post.community.pk)
        exclude_self_query = ~Q(subscriber=post.creator)

        # Exclude banned users
        exclude_blocked_users_query.add(Q(subscriber__banned_of_communities__id=post.community.pk), Q.OR)

        # Subscriptions after excluding blocked users
        target_subscriptions_excluding_blocked = CommunityNotificationsSubscription.objects.filter(
            community_subscriptions_query
            & exclude_self_query  # & community_members_query - Non members can follow posts
        ).exclude(exclude_blocked_users_query)

        staff_members_query = Q(
            subscriber__communities_memberships__community_id=post.community.pk,
            subscriber__communities_memberships__is_administrator=True,
        ) | Q(
            subscriber__communities_memberships__community_id=post.community.pk,
            subscriber__communities_memberships__is_moderator=True,
        )

        # Subscriptions from staff of community
        target_subscriptions_with_staff = CommunityNotificationsSubscription.objects.filter(
            community_subscriptions_query & community_members_query & staff_members_query & exclude_self_query
        )

        results = target_subscriptions_excluding_blocked.union(target_subscriptions_with_staff)

        return results

    @classmethod
    def get_user_notification_target_subscriptions(cls, post):
        UserNotificationsSubscription = get_user_notifications_subscription_model()

        user_subscriptions_query = Q(user=post.creator, new_post_notifications=True)

        exclude_blocked_users_query = Q(
            Q(subscriber__blocked_by_users__blocker_id=post.creator.pk)
            | Q(subscriber__user_blocks__blocked_user_id=post.creator.pk)
        )
        exclude_self_query = ~Q(subscriber=post.creator)

        if post.is_encircled_post():
            circle_ids = [circle.pk for circle in post.circles.all()]
            post_circles_query = Q(subscriber__connections__target_connection__circles__in=circle_ids)
            user_subscriptions_query.add(post_circles_query, Q.AND)

        user_subscriptions_query.add(exclude_self_query, Q.AND)

        # Subscriptions after excluding blocked users
        target_subscriptions = UserNotificationsSubscription.objects.filter(user_subscriptions_query).exclude(
            exclude_blocked_users_query
        )

        return target_subscriptions

    def can_be_viewed(self, user: User) -> bool:
        post = Post.objects.filter(Post.is_visible_query(user)).distinct().get(id=self.id)

        return post is not None

    def count_comments(self):
        return Comment.count_comments_for_commentable(self.pk)

    def count_comments_with_user(self, user):
        # Count comments excluding users blocked by authenticated user
        count_query = ~Q(
            Q(commenter__blocked_by_users__blocker_id=user.pk) | Q(commenter__user_blocks__blocked_user_id=user.pk)
        )

        if self.community:
            if not user.is_staff_of_community_with_name(community_name=self.community.name):
                # Don't retrieve comments except from staff members
                blocked_users_query_staff_members = Q(
                    commenter__communities_memberships__community_id=self.community.pk
                )
                blocked_users_query_staff_members.add(
                    Q(commenter__communities_memberships__is_administrator=True)
                    | Q(commenter__communities_memberships__is_moderator=True),
                    Q.AND,
                )

                count_query.add(~blocked_users_query_staff_members, Q.AND)

        # Don't count soft deleted items
        count_query.add(Q(is_deleted=False), Q.AND)

        return self.comments.filter(count_query).count()

    def count_reactions(self, reactor_id=None):
        return PostReaction.count_reactions_for_post_with_id(self.pk, reactor_id=reactor_id)

    def is_text_only_post(self):
        return self.has_text() and not self.has_video() and not self.has_image()

    def has_text(self):
        return attr_filled(self, "text")

    def has_image(self):
        return attr_filled(self, "image")

    def has_video(self):
        return attr_filled(self, "video")

    def has_links(self):
        return self.links.exists()

    def comment(self, text, commenter, language=None):
        return Comment.create_comment(text=text, commenter=commenter, commentable=self, language=language)

    def comment_on(self, commenter: User, text: str, language: Optional[str]):
        return commenter.comment_post_with_id(post_id=self.pk, text=text, language=language)

    def delete_comment(self, commenter: User, id: uuid.UUID):
        commenter.delete_comment_with_id(post_comment_id=id)

    def react(self, reactor, reaction_type):
        reaction = PostReaction.create_reaction(reactor=reactor, reaction_type=reaction_type, post=self)
        if reactor.id != self.creator.id:
            NotificationsService.send_post_reaction_notification(post=self, reactor=reactor, recipient=self.creator)
        return reaction

    def clear_reaction(self, reactor, reaction_type):
        return PostReaction.remove_reaction(reactor=reactor, reaction_type=reaction_type, post=self)

    def is_public_community_post(self):
        Community = get_community_model()
        return Community.objects.filter(posts__id=self.pk, type=Community.COMMUNITY_TYPE_PUBLIC).exists()

    def is_publicly_visible(self):
        return self.is_public or self.is_public_community_post()

    def is_encircled_post(self):
        return not self.community and not self.is_public

    def update(self, text=None, language=None):
        check_can_be_updated(post=self, text=text)
        self.text = text
        self.language_code = language
        self.is_edited = True
        self.save()

    def get_media(self):
        return self.media

    def add_media(self, file, order=None):
        check_can_add_media(post=self)

        is_in_memory_file = isinstance(file, InMemoryUploadedFile) or isinstance(file, SimpleUploadedFile)

        if is_in_memory_file:
            file_mime = magic.from_buffer(file.read())
        elif isinstance(file, TemporaryUploadedFile):
            file_mime = magic.from_file(file.temporary_file_path())
        else:
            file_mime = magic.from_file(file.name)

        check_mimetype_is_supported_media_mimetypes(file_mime)
        # Mime check moved pointer
        file.seek(0)

        file_mime_types = file_mime.split("/")

        file_mime_type = file_mime_types[0]

        if file_mime_type == "image":
            self._add_media_image(image=file, order=order)
        # Here, we could support video posts. We removed existing okuna code using FFMPEG for encoding. To restore,
        # revert the commit containing this comment. This commit also contains some code for treating GIFs as video.
        else:
            raise ValidationError(_("Unsupported media file type"))

        self.save()

    def get_first_media(self):
        return self.media.first()

    def get_first_media_image(self):
        return self.media.filter(type=PostMedia.MEDIA_TYPE_IMAGE).first()

    def _add_media_image(self, image, order=None):
        return PostImage.create_post_media_image(image=image, post_id=self.pk, order=order)

    def _delete_media_image(self):
        if self.has_image():
            delete_file_field(self.image.image)
            self.image.delete()

    def _update_image(self, image):
        self._delete_media_image()
        self._add_media_image(image=image)

    def count_media(self):
        return self.media.count()

    def publish(self):
        check_can_be_published(post=self)
        if self.has_media() and self.get_media().get().type == PostMedia.MEDIA_TYPE_VIDEO:
            # After finishing, this will call _publish()
            self.status = Post.STATUS_PROCESSING
            self.save()
            process_post_media.delay(post_id=self.pk)
        else:
            self._publish()

    def _publish(self):
        self.status = Post.STATUS_PUBLISHED
        self.created = timezone.now()
        self._process_post_subscribers()
        self.save()

    def is_draft(self):
        return self.status == Post.STATUS_DRAFT

    def is_empty(self):
        return not self.text and not hasattr(self, "image") and not hasattr(self, "video") and not self.has_media()

    def has_media(self):
        return self.media.exists()

    def save(self, *args, **kwargs):
        """On create, update timestamps"""
        if not self.created:
            self.created = timezone.now()

        self.modified = timezone.now()

        super(Post, self).save(*args, **kwargs)

        self._process_post_mentions()
        self._process_post_hashtags()
        self._process_post_links()

        return self

    def delete(self, *args, **kwargs):
        self.delete_media()
        super(Post, self).delete(*args, **kwargs)

    def delete_media(self):
        if self.has_image():
            delete_file_field(self.image.image)

    def soft_delete(self):
        self.delete_notifications()
        for comment in self.comments.all().iterator():
            comment.soft_delete()
        self.is_deleted = True
        self.save()

    def unsoft_delete(self):
        self.is_deleted = False
        for comment in self.comments.all().iterator():
            comment.unsoft_delete()
        self.save()

    def delete_notifications(self):
        # Remove all post reaction notifications
        PostReactionNotification = get_post_reaction_notification_model()
        PostReactionNotification.objects.filter(post_reaction__post_id=self.pk).delete()

        # Remove all post user mention notifications
        PostUserMentionNotification = get_post_user_mention_notification_model()
        PostUserMentionNotification.objects.filter(post_user_mention__post_id=self.pk).delete()

        # Remove all post comment notifications
        PostCommentNotification = get_post_comment_notification_model()
        PostCommentNotification.objects.filter(post_comment__commentable_id=self.pk).delete()

        # Remove all post comment reply notifications
        PostCommentReplyNotification = get_post_comment_reply_notification_model()
        PostCommentReplyNotification.objects.filter(post_comment__commentable_id=self.pk).delete()

        # Remove all post comment reaction notifications
        PostCommentReactionNotification = get_post_comment_reaction_notification_model()
        PostCommentReactionNotification.objects.filter(
            post_comment_reaction__post_comment__commentable_id=self.pk
        ).delete()

        # Remove all post comment user mention notifications
        PostCommentUserMentionNotification = get_post_comment_user_mention_notification_model()
        PostCommentUserMentionNotification.objects.filter(
            post_comment_user_mention__post_comment__commentable_id=self.pk
        ).delete()

        # Remove all community new post notifications
        CommunityNewPostNotification = get_community_new_post_notification_model()
        CommunityNewPostNotification.objects.filter(post_id=self.pk).delete()

        # Remove all user new post notifications
        UserNewPostNotification = get_user_new_post_notification_model()
        UserNewPostNotification.objects.filter(post_id=self.pk).delete()

    def delete_notifications_for_user(self, user):
        # Remove all post reaction notifications
        PostReactionNotification = get_post_reaction_notification_model()
        PostReactionNotification.objects.filter(
            notification__owner_id=user.pk, post_reaction__post_id=self.pk
        ).delete()

        # Remove all post user mention notifications
        PostUserMentionNotification = get_post_user_mention_notification_model()
        PostUserMentionNotification.objects.filter(
            notification__owner_id=user.pk, post_user_mention__post_id=self.pk
        ).delete()

        # Remove all post comment notifications
        PostCommentNotification = get_post_comment_notification_model()
        PostCommentNotification.objects.filter(
            post_comment__commentable_id=self.pk, notification__owner_id=user.pk
        ).delete()

        # Remove all post comment reply notifications
        PostCommentReplyNotification = get_post_comment_reply_notification_model()
        PostCommentReplyNotification.objects.filter(
            post_comment__commentable_id=self.pk, notification__owner_id=user.pk
        ).delete()

        # Remove all post comment user mention notifications
        PostCommentUserMentionNotification = get_post_comment_user_mention_notification_model()
        PostCommentUserMentionNotification.objects.filter(
            notification__owner_id=user.pk,
            post_comment_user_mention__post_comment__commentable_id=self.pk,
        ).delete()

        # Remove all post comment reaction notifications
        PostCommentReactionNotification = get_post_comment_reaction_notification_model()
        PostCommentReactionNotification.objects.filter(
            notification__owner_id=user.pk,
            post_comment_reaction__post_comment__commentable_id=self.pk,
        ).delete()

        # Remove all community new post notifications
        CommunityNewPostNotification = get_community_new_post_notification_model()
        CommunityNewPostNotification.objects.filter(notification__owner_id=user.pk, post_id=self.pk).delete()

        # Remove all user new post notifications
        UserNewPostNotification = get_user_new_post_notification_model()
        UserNewPostNotification.objects.filter(notification__owner_id=user.pk, post_id=self.pk).delete()

    def delete_notifications_except_for_users(self, excluded_users):
        excluded_ids = [user.pk for user in excluded_users]

        # Remove all post user mention notifications
        PostUserMentionNotification = get_post_user_mention_notification_model()
        PostUserMentionNotification.objects.filter(post_user_mention__post_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

        # Remove all post comment notifications
        PostCommentNotification = get_post_comment_notification_model()
        PostCommentNotification.objects.filter(post_comment__commentable_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

        # Remove all post reaction notifications
        PostReactionNotification = get_post_reaction_notification_model()
        PostReactionNotification.objects.filter(post_reaction__post_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

        # Remove all post comment reply notifications
        PostCommentReplyNotification = get_post_comment_reply_notification_model()
        PostCommentReplyNotification.objects.filter(post_comment__commentable_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

        # Remove all post comment reaction notifications
        PostCommentReactionNotification = get_post_comment_reaction_notification_model()
        PostCommentReactionNotification.objects.filter(
            post_comment_reaction__post_comment__commentable_id=self.pk
        ).exclude(notification__owner_id__in=excluded_ids).delete()

        # Remove all post comment user mention notifications
        PostCommentUserMentionNotification = get_post_comment_user_mention_notification_model()
        PostCommentUserMentionNotification.objects.filter(
            post_comment_user_mention__post_comment__commentable_id=self.pk
        ).exclude(notification__owner_id__in=excluded_ids).delete()

        # Remove all community new post notifications
        CommunityNewPostNotification = get_community_new_post_notification_model()
        CommunityNewPostNotification.objects.filter(post_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

        # Remove all user new post notifications
        UserNewPostNotification = get_user_new_post_notification_model()
        UserNewPostNotification.objects.filter(post_id=self.pk).exclude(
            notification__owner_id__in=excluded_ids
        ).delete()

    def _process_post_mentions(self):
        if not self.text:
            self.mentions.clear()
        else:
            usernames = extract_usernames_from_string(string=self.text)
            if not usernames:
                self.mentions.clear()
            else:
                existing_mention_usernames = []
                for existing_mention in self.mentions.all().iterator():
                    existing_mention_username = existing_mention.username.lower()
                    if existing_mention_username not in usernames:
                        existing_mention.delete()
                    else:
                        existing_mention_usernames.append(existing_mention_username)

                PostUserMention = get_post_user_mention_model()
                User = get_user_model()

                for username in usernames:
                    username = username.lower()
                    if username not in existing_mention_usernames:
                        try:
                            user = User.objects.only("id", "username").get(username__iexact=username)
                            user_is_post_creator = user.pk == self.creator_id
                            if not user_is_post_creator:
                                PostUserMention.create_post_user_mention(
                                    user=user,
                                    post=self,
                                    notify=user.can_see_post(post=self),
                                )
                                existing_mention_usernames.append(username)
                        except User.DoesNotExist:
                            pass

    def _process_post_hashtags(self):
        if not self.text:
            self.hashtags.all().delete()
        else:
            hashtags = extract_hashtags_from_string(string=self.text)
            if not hashtags:
                self.hashtags.all().delete()
            else:
                existing_hashtags = []
                for existing_hashtag in self.hashtags.only("id", "name").all().iterator():
                    if existing_hashtag.name not in hashtags:
                        self.hashtags.remove(existing_hashtag)
                    else:
                        existing_hashtags.append(existing_hashtag.name)

                Hashtag = get_hashtag_model()

                for hashtag in hashtags:
                    hashtag = hashtag.lower()
                    hashtag_obj = Hashtag.get_or_create_hashtag(name=hashtag, post=self)
                    if hashtag not in existing_hashtags:
                        self.hashtags.add(hashtag_obj)

    def _process_post_subscribers(self):
        if self.community:
            CommunityNewPostNotification = get_community_new_post_notification_model()
            community_subscriptions = Post.get_community_notification_target_subscriptions(post=self)

            for subscription in community_subscriptions:
                CommunityNewPostNotification.create_community_new_post_notification(
                    post_id=self.pk,
                    owner_id=subscription.subscriber.pk,
                    community_notifications_subscription_id=subscription.pk,
                )
        else:
            UserNewPostNotification = get_user_new_post_notification_model()
            user_subscriptions = Post.get_user_notification_target_subscriptions(post=self)
            for subscription in user_subscriptions:
                UserNewPostNotification.create_user_new_post_notification(
                    post_id=self.pk,
                    owner_id=subscription.subscriber.pk,
                    user_notifications_subscription_id=subscription.pk,
                )

    def _process_post_links(self):
        # always clear, re-add if text contains links. If the text is edited and a link is dropped, this should behave
        # correctly. This leads to completely new entries in the m2m mapping table on every save, but since otherwise,
        # we can't guarantee the order, this is acceptable. This also auto-heals old, wrongly sorted or saved entries on
        # the next save.
        self.links.clear()
        if self.has_text():
            # we need to add all items separately, sequentially for sorting to prevail
            # (instead of setting the whole QuerySet)
            for preview in LinkPreview.parse_text(self.text):
                self.links.add(preview)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.title}"

    @staticmethod
    def does_exist_query(prefix=None):
        pQ = PrefixedQ(prefix)

        return pQ(is_deleted=False)

    @staticmethod
    def is_visible_query(user, prefix=None):
        pQ = PrefixedQ(prefix)

        if user.is_anonymous:
            return pQ(is_deleted=False, status=Post.STATUS_PUBLISHED, is_public=True)

        return pQ(is_deleted=False) & (
            (pQ(creator=user) & (pQ(community__isnull=True) | pQ(community__memberships__user__id=user.id)))
            | (pQ(status=Post.STATUS_PUBLISHED) & (pQ(is_public=True) | pQ(community__memberships__user__id=user.id)))
        )

    @property
    def comment_notification_context(self) -> CommentNotificationContext:
        return CommentNotificationContext(id=self.id, title=self.title, context=self.community.title)


class TopPost(ModelWithUUID):
    post = models.OneToOneField(Post, on_delete=models.CASCADE, related_name="top_post")
    created = models.DateTimeField(editable=False, db_index=True)

    def save(self, *args, **kwargs):
        """On save, update timestamps"""
        if not self.created:
            self.created = timezone.now()
        self.modified = timezone.now()

        return super(TopPost, self).save(*args, **kwargs)


class TrendingPost(ModelWithUUID):
    post = models.OneToOneField(Post, on_delete=models.CASCADE, related_name="trending_post")
    created = models.DateTimeField(editable=False, db_index=True)

    def save(self, *args, **kwargs):
        """On save, update timestamps"""
        if not self.created:
            self.created = timezone.now()
        self.modified = timezone.now()

        return super(TrendingPost, self).save(*args, **kwargs)


class TopPostCommunityExclusion(ModelWithUUID):
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="top_posts_community_exclusions")
    community = models.ForeignKey(
        "openbook_communities.Community",
        on_delete=models.CASCADE,
        related_name="top_posts_community_exclusions",
    )
    created = models.DateTimeField(editable=False, db_index=True)

    class Meta:
        unique_together = (
            "user",
            "community",
        )

    def save(self, *args, **kwargs):
        """On save, update timestamps"""
        if not self.created:
            self.created = timezone.now()
        self.modified = timezone.now()

        return super(TopPostCommunityExclusion, self).save(*args, **kwargs)


class ProfilePostsCommunityExclusion(ModelWithUUID):
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="profile_posts_community_exclusions",
    )
    community = models.ForeignKey(
        "openbook_communities.Community",
        on_delete=models.CASCADE,
        related_name="profile_posts_community_exclusions",
    )
    created = models.DateTimeField(editable=False, db_index=True)

    class Meta:
        unique_together = (
            "user",
            "community",
        )

    def save(self, *args, **kwargs):
        """On save, update timestamps"""
        if not self.created:
            self.created = timezone.now()
        self.modified = timezone.now()

        return super(ProfilePostsCommunityExclusion, self).save(*args, **kwargs)


class PostMedia(OrderedModelWithUUID):
    post = models.ForeignKey(Post, on_delete=models.CASCADE, related_name="media")

    MEDIA_TYPE_VIDEO = "V"
    MEDIA_TYPE_IMAGE = "I"

    MEDIA_TYPES = (
        (MEDIA_TYPE_VIDEO, "Video"),
        (MEDIA_TYPE_IMAGE, "Image"),
    )

    type = models.CharField(max_length=5, choices=MEDIA_TYPES)

    # Generic relation types
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.UUIDField()
    content_object = GenericForeignKey()

    @classmethod
    def create_post_media(cls, post_id, type, content_object, order):
        return cls.objects.create(type=type, content_object=content_object, post_id=post_id, order=order)


class PostImage(ModelWithUUID):
    post = models.OneToOneField(Post, on_delete=models.CASCADE, related_name="image", null=True)
    image = ProcessedImageField(
        verbose_name=_("image"),
        storage=post_image_storage,
        upload_to=upload_to_post_image_directory,
        width_field="width",
        height_field="height",
        blank=False,
        null=True,
        processors=[
            ExifRotate(),
            ResizeToFit(
                width=settings.CONTENT_IMAGE_MAX_DIMENSION,
                height=settings.CONTENT_IMAGE_MAX_DIMENSION,
                upscale=False,
            ),
        ],
        max_length=settings.IMAGE_URL_MAX_LENGTH,
    )
    image_blurhash = models.CharField(max_length=50, blank=True, null=True)
    width = models.PositiveIntegerField(editable=False, null=False, blank=False)
    height = models.PositiveIntegerField(editable=False, null=False, blank=False)
    media = GenericRelation(PostMedia)

    @classmethod
    def create_post_media_image(cls, image, post_id, order):
        post_image = cls.objects.create(image=image, post_id=post_id, image_blurhash=generate_blurhash(image))
        PostMedia.create_post_media(
            type=PostMedia.MEDIA_TYPE_IMAGE,
            content_object=post_image,
            post_id=post_id,
            order=order,
        )
        return post_image


class PostReaction(ModelWithUUID):
    post = models.ForeignKey(Post, on_delete=models.CASCADE, related_name="reactions")
    created = models.DateTimeField(editable=False)
    reactor = models.ForeignKey(User, on_delete=models.CASCADE, related_name="post_reactions")
    reaction_type = models.CharField(max_length=20, choices=ReactionType.choices)

    class Meta:
        unique_together = (
            "reactor",
            "post",
        )

    @classmethod
    def create_reaction(cls, reactor, reaction_type, post):
        PostReaction.objects.filter(reactor=reactor, post=post).delete()
        return PostReaction.objects.create(reactor=reactor, reaction_type=reaction_type, post=post)

    @classmethod
    def remove_reaction(cls, reactor, reaction_type, post):
        return PostReaction.objects.filter(reactor=reactor, reaction_type=reaction_type, post=post).delete()

    @classmethod
    def count_reactions_for_post_with_id(cls, post_id, reactor_id=None, reaction_type=None):
        count_query = Q(post_id=post_id, reactor__is_deleted=False)

        if reactor_id:
            count_query.add(Q(reactor_id=reactor_id), Q.AND)

        if reaction_type:
            count_query.add(Q(reaction_type=reaction_type), Q.AND)

        return cls.objects.filter(count_query).count()

    def save(self, *args, **kwargs):
        """On save, update timestamps"""
        if not self.created:
            self.created = timezone.now()
        return super(PostReaction, self).save(*args, **kwargs)


class PostMute(ModelWithUUID):
    post = models.ForeignKey(Post, on_delete=models.CASCADE, related_name="mutes")
    muter = models.ForeignKey(User, on_delete=models.CASCADE, related_name="post_mutes")

    class Meta:
        unique_together = (
            "post",
            "muter",
        )

    @classmethod
    def create_post_mute(cls, post_id, muter_id):
        return cls.objects.create(post_id=post_id, muter_id=muter_id)


class PostUserMention(ModelWithUUID):
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="post_mentions")
    post = models.ForeignKey(Post, on_delete=models.CASCADE, related_name="user_mentions")

    class Meta:
        unique_together = (
            "user",
            "post",
        )

    @classmethod
    def create_post_user_mention(cls, user, post, notify=True):
        post_user_mention = cls.objects.create(user=user, post=post)
        if notify:
            PostUserMentionNotification = get_post_user_mention_notification_model()
            PostUserMentionNotification.create_post_user_mention_notification(
                post_user_mention_id=post_user_mention.pk, owner_id=user.pk
            )
        return post_user_mention
